# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import logging, sys
LOG = logging.getLogger(__name__)

from manifesttool import create
from urllib3.exceptions import MaxRetryError
import copy
import time
import tempfile
import os
import os.path
import shutil

STOP_STATES = {'autostopped',
               'conflict',
               'expired',
               'manifestremoved',
               'quotaallocationfailed',
               'userstopped'}

MAX_NAME_LEN = 128 # The update API has a maximum name length of 128, but this is not queriable.

def main(options):
    try:
        # import mbed_cloud.update
        from mbed_cloud.update import UpdateAPI
        # from mbed_cloud.device_directory import DeviceDirectoryAPI
        import mbed_cloud.exceptions
    except:
        LOG.critical('manifest-tool update commands require installation of the Mbed Cloud SDK:'
                     ' https://github.com/ARMmbed/mbed-cloud-sdk-python')
        return 1

    LOG.debug('Preparing an update on Mbed Cloud')
    # upload a firmware
    api = None
    # dd_api = None
    try:
        # If set use api key set in manifest-tool update.
        if hasattr(options, 'api_key') and options.api_key:
            tempKey = options.api_key
            config = {'api_key': tempKey}
            api = UpdateAPI(config)
        # Otherwise use API key set in manifest-tool init
        else: api = UpdateAPI()
        # dd_api = DeviceDirectoryAPI()

    except ValueError:
        LOG.critical('API key is required to connect to the Update Service. It can be added using manifest-tool init -a'
                     ' <api key> or by manually editing .mbed_cloud_config.json')
        return 1
    if not options.payload_name:
        name = os.path.basename(options.payload.name) + time.strftime('-%Y-%m-%dT%H:%M:%S')
        LOG.info('Using {} as payload name.'.format(name))
        options.payload_name = name
    if len(options.payload_name) > MAX_NAME_LEN:
        LOG.critical(
            'Payload name is too long. Maximum length is {size}. ("{base}" <- {size}. overflow -> "{overflow}")'.format(
                size=MAX_NAME_LEN, base=options.payload_name[:MAX_NAME_LEN],
                overflow=options.payload_name[MAX_NAME_LEN:]))
        return 1

    if not options.manifest_name:
        name = os.path.basename(options.payload.name) + time.strftime('-%Y-%m-%dT%H:%M:%S-manifest')
        LOG.info('Using {} as manifest name.'.format(name))
        options.manifest_name = name

    if len(options.manifest_name) > MAX_NAME_LEN:
        LOG.critical(
            'Manifest name is too long. Maximum length is {size}. ("{base}" <- {size}. overflow -> "{overflow}")'.format(
                size=MAX_NAME_LEN, base=options.manifest_name[:MAX_NAME_LEN],
                overflow=options.manifest_name[MAX_NAME_LEN:]))
        return 1
    campaign_name = options.payload.name + time.strftime('-%Y-%m-%dT%H:%M:%S-campaign')
    if len(campaign_name) > MAX_NAME_LEN:
        LOG.critical(
            'Campaign name is too long. Maximum length is {size}. ("{base}" <- {size}. overflow -> "{overflow}")'.format(
                size=MAX_NAME_LEN, base=campaign_name[:MAX_NAME_LEN],
                overflow=campaign_name[MAX_NAME_LEN:]))
        return 1
    query_name = options.payload.name + time.strftime('-%Y-%m-%dT%H:%M:%S-filter')
    if len(query_name) > MAX_NAME_LEN:
        LOG.critical(
            'Filter name is too long. Maximum length is {size}. ("{base}" <- {size}. overflow -> "{overflow}")'.format(
                size=MAX_NAME_LEN, base=query_name[:MAX_NAME_LEN],
                overflow=query_name[MAX_NAME_LEN:]))
        return 1


    payload = None
    manifest = None
    campaign = None
    RC = 0
    handled = False
    manifest_file = None
    tempdirname = tempfile.mkdtemp()
    try:
        kwArgs = {}
        if options.payload_description:
            kwArgs['description'] = options.payload_description
        try:
            payload = api.add_firmware_image(
                            name = options.payload_name,
                        datafile = options.payload.name,
                        **kwArgs)
        except mbed_cloud.exceptions.CloudApiException as e:
            # TODO: Produce a better failuer message
            LOG.critical('Upload of payload failed with:\n{}'.format(e).rstrip())
            handled = True
            LOG.critical('Check API server URL "{}"'.format(api.config["host"]))
            raise e
        except MaxRetryError as e:
            LOG.critical('Upload of payload failed with:\n{}'.format(e))
            handled=True
            LOG.critical('Failed to establish connection to API-GW')
            LOG.critical('Check API server URL "{}"'.format(api.config["host"]))
            raise e

        LOG.info("Created new firmware at {}".format(payload.url))
        options.payload.seek(0)
        # create a manifest
        create_opts = copy.copy(options)
        create_opts.uri = payload.url
        create_opts.payload = options.payload
        if not (hasattr(create_opts, "output_file") and create_opts.output_file):
            try:
                manifest_file = open(os.path.join(tempdirname,'manifest'),'wb')
                LOG.info("Created temporary manifest file at {}".format(manifest_file.name))
                create_opts.output_file = manifest_file
            except IOError as e:
                LOG.critical("Failed to create temporary manifest file with:")
                print(e)
                LOG.critical("Try using '-o' to output a manifest file at a writable location.")
                handled = True
                raise e
        try:
            rc = create.main(create_opts)
        except IOError as e:
            LOG.critical("Failed to create manifest with:")
            print(e)
            handled = True
            raise e

        if rc:
            return rc

        kwArgs = {}
        if options.manifest_description:
            kwArgs['description'] = options.manifest_description

        manifest_path = create_opts.output_file.name
        create_opts.output_file.close()
        # upload a manifest
        try:
            manifest = api.add_firmware_manifest(
                            name = options.manifest_name,
                        datafile = manifest_path,
                        **kwArgs)
        except mbed_cloud.exceptions.CloudApiException as e:
            # TODO: Produce a better failure message
            LOG.critical('Upload of manifest failed with:')
            print(e)
            LOG.critical("Try using '-o' to output a manifest file at a writable location.")
            handled = True
            raise e

        LOG.info('Created new manifest at {}'.format(manifest.url))
        LOG.info('Manifest ID: {}'.format(manifest.id))

        try:
            campaign = api.add_campaign(
                name = campaign_name,
                manifest_id = manifest.id,
                device_filter = {'id': { '$eq': options.device_id }},
            )
        except mbed_cloud.exceptions.CloudApiException as e:
            LOG.critical('Campaign creation failed with:')
            print(e)
            handled = True
            raise e

        LOG.info('Campaign successfully created. Current state: %r' % (campaign.state))
        LOG.info('Campaign successfully created. Filter result: %r' % (campaign.device_filter))

        LOG.info("Starting the update campign...")

        # By default a new campaign is created with the 'draft' status. We can manually start it.
        try:
            new_campaign = api.start_campaign(campaign)
            new_campaign = None
        except mbed_cloud.exceptions.CloudApiException as e:
            LOG.critical('Starting campaign failed with:')
            print(e)
            handled = True
            raise e

        LOG.info("Campaign successfully started. Current state: %r. Checking updates.." % (campaign.state))
        oldstate = api.get_campaign(campaign.id).state
        LOG.info("Current state: %r" % (oldstate))
        timeout = options.timeout
        while timeout != 0:
            c = api.get_campaign(campaign.id)
            if oldstate != c.state:
                LOG.info("Current state: %r" % (c.state))
                oldstate = c.state
            if c.state in STOP_STATES:
                LOG.info("Finished in state: %r" % (c.state))
                break
            time.sleep(1)
            if timeout > 0:
                timeout -= 1
        if timeout == 0:
            LOG.critical("Campaign timed out")
            RC = 1

    except KeyboardInterrupt as e:
        LOG.critical('User Aborted... Cleaning up.')
        RC = 1
    except:
        if not handled:
            LOG.critical('Unhandled Exception:')
            import traceback
            traceback.print_exc(file=sys.stdout)
        RC = 1
    finally:
        # cleanup
        if manifest_file:
            manifest_file.close()
        shutil.rmtree(tempdirname)
        if not options.no_cleanup:
            LOG.info("** Deleting update campaign and manifest **")
            try:
                if campaign and campaign.id:
                    api.delete_campaign(campaign.id)
                if manifest and manifest.id:
                    api.delete_firmware_manifest(manifest.id)
                if payload and payload.id:
                    api.delete_firmware_image(payload.id)
                # dd_api.delete_query(new_query.id)
            except mbed_cloud.exceptions.CloudApiException as e:
                LOG.critical('Cleanup of campaign failed with:')
                print(e)
                RC = 1
    return RC



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import json, logging, binascii
from manifesttool import codec
from manifesttool.v1 import manifest_definition

# All supported decoders
from pyasn1.codec.der import decoder as der_decoder

LOG = logging.getLogger(__name__)

# Wrapper class with same signature as pyasn1 decoders
class json_decoder(object):
    def decode(self, data, schema):
        return (json.dumps(data),)

def parse(options):
    data = options.input_file.read()
    LOG.debug('Read {} bytes of encoded data. Will try to decode...'.format(len(data)))

    decoded_data = {
        "der": lambda d: codec.bin2obj(d, manifest_definition.SignedResource(), der_decoder, True),
    }[options.encoding](data)
    LOG.debug('Successfully decoded data from {} encoded binary'.format(options.encoding))
    return decoded_data



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

# This file is a python class representation of the ASN.1 schema defined
# in manifest.asn. It is used to generate a JSON representation of the manifest
# from raw data.
#
# It differs from manifest_definition.py as it's not used by pyasn1, it's not used
# to encode the manifest data nor is it automatically generated.
#
# The process looks like:
#   (Input data) -> (Python object representation) -> (JSON) -> (PyASN1 object) -> ([DER] encoding)
#                             [ this file ]
import time, collections
from manifesttool import utils
from manifesttool.errorhandler import InvalidObject

class BaseObject(object):

    # Basic check to see if all required fields are set.
    # Does NOT check if the set value is of right type accoridng to ASN.1 definition
    def verify(self, required_fields):
        for f in required_fields:
            if self.__dict__.get(f) is None:
                raise InvalidObject("Value for field %r not set, but it's required" % f)

    # Convinience method to just convert object to python dict
    def to_dict(self):
        return utils.todict(self)

class Resource(BaseObject):
    TYPE_MANIFEST, TYPE_FIRMWARE = range(2)

    def __init__(self, resource, resourceType, uri=None):
        self.uri = uri
        self.resourceType = resourceType
        self.resource = resource

        # If type is manifest, then key is manifest. payload if not.
        k = {
            Resource.TYPE_MANIFEST: "manifest",
            Resource.TYPE_FIRMWARE: "payload"
        }[self.resourceType]
        self.resource = {
            k: resource
        }

class Manifest(BaseObject):
    # Manifest version enum values
    V1 = 1

    # Encryption mode enum values
    ENCRYPTION_MODE_INVALID = 0
    ENCRYPTION_MODE_AES_128 = 1
    ENCRYPTION_MODE_NONE_ECC = 2
    ENCRYPTION_MODE_NONE = 3
    ENCRYPTION_MODE_NONE_PSK_CCM_16 = 4
    ENCRYPTION_MODE_PSK_CCM_16 = 4


    def __init__(self, **kwargs):
        # Set defaults
        self.manifestVersion = Manifest.V1
        self.timestamp = int(time.time())
        self.deviceId = None
        self.vendorId = None
        self.classId = None
        self.vendorInfo = None
        self.description = None
        self.nonce = None
        self.applyImmediately = False
        # self.applyPeriod = {
        #     'validFrom': int(time.time()),
        #     'validTo': int(time.time()) + (3600 * 24 * 30)
        # }
        self.applyPeriod = None
        self.encryptionMode = {
            "enum" : Manifest.ENCRYPTION_MODE_NONE
        }
        self.aliases = []
        self.dependencies = []
        self.payload = None

        # Override with kwargs
        for key, value in kwargs.items():
            if value is None:
                continue
            if hasattr(self, key):
                setattr(self, key, value)

        # Check that all required fields are set
        self.verify(["manifestVersion", "timestamp", "deviceId", "vendorId", "classId", "vendorInfo",\
                     "applyImmediately", "dependencies", "nonce", "aliases"])

class PayloadDescription(BaseObject):
    FORMAT_UNDEFINED = 0
    FORMAT_RAW_BINARY = 1
    FORMAT_CBOR = 2
    FORMAT_HEX_LOCATION_LENGTH_DATA = 3
    FORMAT_ELF = 4

    def __init__(self, **kwargs):
        # Set defaults
        self.format = {
            "enum": PayloadDescription.FORMAT_RAW_BINARY
        }
        self.encryptionInfo = None
        self.storageIdentifier = None
        self.reference = None
        self.version = None

        # Override with kwargs
        for key, value in kwargs.items():
            if value is None:
                continue
            if hasattr(self, key):
                setattr(self, key, value)

        # Verify that all required fields are set
        self.verify(["format", "storageIdentifier", "reference"])

class SignedResource(BaseObject):
    def __init__(self, resource, signature):
        self.resource = resource
        self.signature = signature

class ResourceReference(object):
    def __init__(self, hash, size, uri=None):
        self.hash = hash
        self.uri = uri
        self.size = size

class ResourceAlias(object):
    def __init__(self, hash, uri):
        self.hash = hash
        self.uri = uri

class CertificateReference(object):
    def __init__(self, fingerprint, uri):
        self.fingerprint = fingerprint
        self.uri = uri

class SignatureBlock(object):
    def __init__(self, signature, certificates):
        self.signature = signature
        self.certificates = certificates

class MacBlock(object):
    def __init__(self, pskID, keyTableIV, keyTableVersion, keyTableIndexSize, keyTableRecordSize, keyTableRef=None):
        self.pskID = pskID
        self.keyTableIV = keyTableIV
        self.keyTableVersion = keyTableVersion
        self.keyTableRef = keyTableRef
        self.keyTableIndexSize = keyTableIndexSize
        self.keyTableRecordSize = keyTableRecordSize

class ResourceSignature(object):
    def __init__(self, hash, signatures, macs=None):
        self.hash = hash
        self.signatures = signatures
        if macs:
            self.macs = macs
